#!/usr/bin/env python
import re
import argparse
import pandas as pd
from SLRanger.run_ex_function import run_track_cluster
# 解析GFF文件并构建DataFrame
def parse_gff(gff_file):
    genes = []
    with open(gff_file, 'r') as file:
        for line in file:
            if line.startswith("#") or line.strip() == '':
                continue
            parts = line.strip().split('\t')
            if parts[2] == 'gene':
                attr_field = parts[8]
                gene_id = None
                for attr in attr_field.split(';'):
                    if attr.startswith('ID='):
                        gene_id = attr.split('ID=')[1].split(';')[0]
                genes.append({
                    'gene': gene_id,
                    'chromosome': parts[0],
                    'start': int(parts[3]),
                    'end': int(parts[4]),
                    'strand': parts[6]
                })
    return pd.DataFrame(genes)

def parse_cds_gene(gff_file):
    """
    解析 GFF 文件，查找哪些 gene 具有 CDS。
    """
    gene_map = {}  # 存储 transcript -> gene 的映射
    cds_parents = set()  # 存储有 CDS 的 transcript ID
    genes_with_cds = set()  # 存储最终有 CDS 的 gene ID

    with open(gff_file, 'r') as f:
        for line in f:
            if line.startswith("#"):
                continue  # 跳过注释行

            fields = line.strip().split("\t")
            if len(fields) < 9:
                continue  # 确保是完整的 GFF 行

            feature_type = fields[2]
            attributes = fields[8]

            if feature_type in ['transcript', 'mRNA']:
                # 提取 transcript ID 和其对应的 gene ID
                transcript_match = re.search(r'ID=([^;]+)', attributes)
                parent_match = re.search(r'Parent=([^;]+)', attributes)
                if transcript_match and parent_match:
                    transcript_id = transcript_match.group(1)
                    gene_id = parent_match.group(1)
                    gene_map[transcript_id] = gene_id

            elif feature_type == "CDS":
                # 提取 CDS 的 Parent（即 transcript）
                cds_match = re.search(r'Parent=([^;]+)', attributes)
                if cds_match:
                    transcript_id = cds_match.group(1).split(',')[0]
                    cds_parents.add(transcript_id)

    # 通过 transcript 找 gene
    for transcript_id in cds_parents:
        if transcript_id in gene_map:
            genes_with_cds.add(gene_map[transcript_id])

    # 转换为 DataFrame 输出
    df = pd.DataFrame({"gene": list(genes_with_cds)})
    return df

# 排序基因并计算基因之间的距离
def sort_and_calc_distance(df):
    df_pos_list = []
    df_neg_list = []

    for chrom in df['chromosome'].unique():
        # 正链
        df_pos_chrom = df[(df['chromosome'] == chrom) & (df['strand'] == '+')].sort_values('end').reset_index(
            drop=True)
        df_pos_chrom['rank'] = range(1, len(df_pos_chrom) + 1)
        df_pos_chrom['intergenic_distance'] = df_pos_chrom['start'] - df_pos_chrom['end'].shift(1)
        df_pos_list.append(df_pos_chrom)

        # 负链
        df_neg_chrom = df[(df['chromosome'] == chrom) & (df['strand'] == '-')].sort_values('end',
                                                                                           ascending=False).reset_index(
            drop=True)
        df_neg_chrom['intergenic_distance'] = df_neg_chrom['start'].shift(1) - df_neg_chrom['end']
        df_neg_chrom['rank'] = range(1, len(df_neg_chrom) + 1)
        df_neg_list.append(df_neg_chrom)

    df_pos_final = pd.concat(df_pos_list).reset_index(drop=True)
    df_neg_final = pd.concat(df_neg_list).reset_index(drop=True)

    df_pos_final['rank'] = df_pos_final.groupby('chromosome').cumcount() + 1
    df_neg_final['rank'] = df_neg_final.groupby('chromosome').cumcount() + 1
    df_pos_final['strand'] = '+'
    df_neg_final['strand'] = '-'

    df_final = pd.concat([df_pos_final, df_neg_final])
    df_final_s = df_final[['gene', 'strand', 'chromosome', 'rank', 'intergenic_distance']]
    return df_final_s

def sw_ratio(df, cols):
    df_long = pd.melt(df, value_vars=cols, var_name='group', value_name='score')
    df_counts = df_long.groupby(['score', 'group']).size().reset_index(name='count')
    # len_dict={
    #     'random':df_long[df_long['group']=='random'].shape[0],
    #     'sw':df_long[df_long['group']=='sw'].shape[0]
    # }
    # df_counts['count'] = df_counts.apply(lambda x: len_dict[x['group']]-x['count'], axis=1)
    df_wide = df_counts.pivot(index='score', columns='group', values='count').fillna(0)

    df_wide['ratio'] = df_wide[cols[1]] / df_wide[cols[0]]
    return df_wide

def cutoff(data, cf):
    df_wide_sw = sw_ratio(data, ['random', 'sw'])
    df_wide_sw.reset_index(inplace=True)
    # sw_sum = df_wide_sw['sw'][df_wide_sw['ratio'] > cf].sum()
    sw_min = df_wide_sw['score'][df_wide_sw['ratio'] > cf].min()
    return sw_min

def sl_process(path, cf):
    sl = pd.read_csv(path, sep='\t')
    sl = sl.dropna()
    sl['SL_score'] = sl['SL_score'].astype(float)
    sl['random'] = (sl['random_SL_score'] * 2).round() / 2
    sl['sw'] = (sl['SL_score'] * 2).round() / 2
    sl_s = sl[sl['SL_score'] > cutoff(sl, cf)]
    sl_s = sl_s[(sl_s['SL_type'] != 'random') & (sl_s['SL_type'] != 'SL1_unknown')]
    sl_s['SL'] = sl_s['SL_type'].apply(lambda x: 'SL1' if x == 'SL1' else 'SL2')
    return sl_s[['query_name', 'SL']]

def fusion_expand(df, genes_dict):
    result_df = pd.DataFrame(columns=df.columns)
    df_with_or = df[df['gene'].str.contains(';', regex=True)]
    df_without_or = df[~df['gene'].str.contains(';', regex=True)]
    # 处理每一行
    for index, row in df_with_or.iterrows():
        gene = row['gene']
        split_genes = gene.split(";")
        temp_dict = {'gene': [], 'strand': [], 'chromosome': [],  'start': [], 'end': []}
        for gene in split_genes:
            if gene in genes_dict:
                temp_dict['gene'].append(gene)
                temp_dict['strand'].append(genes_dict[gene]['strand'])
                temp_dict['chromosome'].append(genes_dict[gene]['chromosome'])
                temp_dict['start'].append(genes_dict[gene]['start'])
                temp_dict['end'].append(genes_dict[gene]['end'])
        new_row = row.copy()
        if len(temp_dict['gene']) > 1:
            all_plus = all(info == '+' for info in temp_dict['strand'])
            all_minus = all(info == '-' for info in temp_dict['strand'])

            if all_plus:
                min_start = min(temp_dict['start'])
                min_start_index = temp_dict['start'].index(min_start)  # x[1] 是 start
                min_start_gene = temp_dict['gene'][min_start_index]
                min_end = temp_dict['end'][min_start_index]
                max_start = min_start + (min_end - min_start)/2
                new_row['gene'] = min_start_gene
                result_df = pd.concat([result_df, pd.DataFrame([new_row])], ignore_index=True)
                for start in temp_dict['start']:
                    if min_start < start <= max_start:  # 与最小 start 差值不超过 300bp
                        start_index = temp_dict['start'].index(start)
                        new_row['gene'] = temp_dict['gene'][start_index]
                        result_df = pd.concat([result_df, pd.DataFrame([new_row])], ignore_index=True)
            elif all_minus:
                max_end = max(temp_dict['end'])
                max_end_index = temp_dict['end'].index(max_end)  # x[1] 是 start
                max_end_gene = temp_dict['gene'][max_end_index]
                max_start = temp_dict['start'][max_end_index]
                min_end = max_end - (max_end - max_start)/2
                new_row['gene'] = max_end_gene
                result_df = pd.concat([result_df, pd.DataFrame([new_row])], ignore_index=True)
                for end in temp_dict['end']:
                    if min_end < end <= max_end:  # 与最小 start 差值不超过 300bp
                        end_index = temp_dict['end'].index(end)
                        new_row['gene'] = temp_dict['gene'][end_index]
                        result_df = pd.concat([result_df, pd.DataFrame([new_row])], ignore_index=True)
        elif len(temp_dict['gene']) == 1:
            new_row['gene'] = temp_dict['gene'][0]
            result_df = pd.concat([result_df, pd.DataFrame([new_row])], ignore_index=True)

    result_df = pd.concat([df_without_or, result_df], ignore_index=True)
    return result_df

def reshape(df_input):
    gene_data = {}
    for index, row in df_input.iterrows():
        gene = row['gene']  # 获取基因ID
        sl_type = row['SL']  # 获取SL类型
        count = int(row['count'])  # 获取数量

        # 如果基因还不在字典中，初始化
        if gene not in gene_data:
            gene_data[gene] = {'SL1': 0, 'SL2': 0}

        # 更新相应的SL计数
        gene_data[gene][sl_type] = count
    df = pd.DataFrame.from_dict(gene_data, orient='index').reset_index()
    df.columns = ['gene', 'SL1', 'SL2']
    return df

def operon_ref_process(path):
    operon = pd.read_csv(path, sep='\t', header=None)
    operon_s = operon[[8]]
    operon_s.columns = ['attributes']
    expanded_rows = []
    gene_add_row = []
    for row in operon_s['attributes']:
        parts = row.split(';')
        name = parts[0].split('=')[1]
        genes = parts[1].split('=')[1]
        gene_list = genes.split(',')
        gene_all = []
        for gene in gene_list:
            gene2 = 'Gene:' + gene
            expanded_rows.append({'operon': name, 'gene': gene2})
            gene_all.append(gene2)
        gene_add_row.append({'operon': name, 'gene': ",".join(gene_all)})
    gene_add_df = pd.DataFrame(gene_add_row)
    gene_add_df['gene'] = gene_add_df['gene'].apply(lambda x: ','.join(sorted(x.split(','))))

    return gene_add_df, pd.DataFrame(expanded_rows)

def extract_operon_names(df, median_value_sl2):
    operons = []
    i = 0
    n = df.shape[0]
    while i < n:
        if df.loc[i, 'type2'] == 'SL2':
            operon_genes = []
            sum_counts = 0
            if i > 0:
                if df.loc[i - 1, 'type'] == 'SL1' and df.loc[i - 1, 'SL1'] > 1:
                    operon_genes.append(df.loc[i - 1, 'gene'])
                elif (pd.isna(df.loc[i - 1, 'type']) and df.loc[i, 'type'] == 'SL2'
                      and df.loc[i, 'SL2'] >= median_value_sl2):
                    operon_genes.append(df.loc[i - 1, 'gene'])
            operon_genes.append(df.loc[i, 'gene'])
            sum_counts += df.loc[i, 'sum_count']
            i += 1

            while i < n and df.loc[i, 'type2'] == 'SL2':
                operon_genes.append(df.loc[i, 'gene'])
                sum_counts += df.loc[i, 'sum_count']
                i += 1
            # 将得到的operon_genes加入operons列表
            if len(operon_genes) > 1 and sum_counts >= 3:
                operons.append(operon_genes)
            elif len(operon_genes) == 1 and df.loc[i-1, 'type'] == 'SL2' and sum_counts >= median_value_sl2:
                operons.append(operon_genes)
        else:
            i += 1

    return operons

def group_genes_into_operons(df, distance, median_value_sl2):
    operons = []
    current_operon = []

    for idx, row in df.iterrows():
        if pd.isna(row['intergenic_distance']) or row['intergenic_distance'] > distance:
            # 当遇到距离>5000的基因时，判断并保存之前的operon
            if current_operon:
                operons.append(current_operon)
                current_operon = []
        current_operon.append(row)
    # 保存最后一个operon
    if current_operon:
        operons.append(current_operon)

    # 对operon内的基因进行SL判断并重组
    operon_list = []
    for operon in operons:
        operon_df = pd.DataFrame(operon)

        sl2_genes = operon_df[operon_df['type2'] == 'SL2']['gene'].tolist()

        if sl2_genes and len(operon_df) > 1:
            operon_df = operon_df.reset_index(drop=True)
            # value = operon_df['rank'].iloc[0]
            # chr= operon_df['chromosome'].iloc[0]
            operons_t = extract_operon_names(operon_df, median_value_sl2)

            if operons_t:
                operon_list.extend(operons_t)

    dup_operon = list(set(tuple(sublist) for sublist in operon_list))
    return dup_operon


def merge_single_gene_sublists(gene_list, gene_df):
    modified_list = list(set(gene_list))
    # 找到所有只有一个基因的子列表
    single_gene_sublists = [sublist for sublist in gene_list if len(sublist) == 1]
    multiple_gene_sublists = [sublist for sublist in gene_list if len(sublist) > 1]
    gene_to_sublists = {}
    for sublist in multiple_gene_sublists:
        for gene in sublist:
            gene_to_sublists[gene] = sublist  # 每个基因都映射到它所在的子列表

    # 遍历单基因子列表
    for sublist in single_gene_sublists:
        gene = sublist[0]
        gene_row = gene_df[gene_df['gene'] == gene]
        chromosome, strand, rank = gene_row['chromosome'].tolist()[0], gene_row['strand'].tolist()[0], gene_row['rank'].tolist()[0]
        rank_minus_1 = rank - 1
        rank_plus_1 = rank + 1
        filtered_minus = gene_df[(gene_df['chromosome'] == chromosome) &
                                 (gene_df['strand'] == strand) &
                                 (gene_df['rank'] == rank_minus_1)]
        rank_minus_gene = filtered_minus['gene'].tolist()[0] if not filtered_minus.empty else None

        filtered_plus = gene_df[(gene_df['chromosome'] == chromosome) &
                                (gene_df['strand'] == strand) &
                                (gene_df['rank'] == rank_plus_1)]
        rank_plus_gene = filtered_plus['gene'].tolist()[0] if not filtered_plus.empty else None
        target_sublist = []
        if rank_minus_gene in gene_to_sublists:
            target_sublist = list(gene_to_sublists[rank_minus_gene])
            target_sublist.append(gene)
        elif rank_plus_gene in gene_to_sublists:
            target_sublist = list(gene_to_sublists[rank_plus_gene])
            target_sublist.insert(0, gene)

        if target_sublist:
            modified_list.append(target_sublist)
    # 去掉是另一个子列表子集的列表
    filtered_sublists = []
    for sublist in modified_list:
        if not any(set(sublist).issubset(set(other)) and sublist != other for other in modified_list):
            filtered_sublists.append(sublist)

    # 仅返回被修改过的子列表 + 没有被合并的单基因子列表
    return filtered_sublists

def generate_operon_gff(gene_combinations, gene_dict):
    """
    gene_combinations: list of lists, 每个子list包含一组gene ID
    gene_dict: dict, key是gene ID, value是包含start和end的dict/tuple
    chromosome: 染色体编号，默认"I"
    """
    operon_gff_dict = {'chromosome': [], 'source': [], 'type':[], 'start': [], 'end': [], 'score': [],
                       'strand': [], 'phase': [],'gene': []}
    gene_gff_dict = {'chromosome': [], 'source': [], 'type':[], 'start': [], 'end': [], 'score': [],
                       'strand': [], 'phase': [],'gene': []}
    # 处理每个gene组合
    for idx, genes in enumerate(gene_combinations):
        if not genes:  # 跳过空组合
            continue
        formatted_idx = f"LRS{idx + 1:04d}"   # idx + 1 因为要从 0001 开始，04d 确保 4 位补零
        # 获取所有gene的start和end
        starts = []
        ends = []
        chromosomes = []
        strands = []  # 检查方向是否一致

        for gene in genes:
            if gene in gene_dict:
                chromosome = gene_dict[gene]['chromosome']
                start = gene_dict[gene]['start']
                end = gene_dict[gene]['end']
                strand = gene_dict[gene]['strand']

                gene_gff_dict['chromosome'].append(chromosome)
                gene_gff_dict['strand'].append(strand)
                gene_gff_dict['start'].append(start)
                gene_gff_dict['end'].append(end)
                gene_entry = f"ID={gene};Parent={formatted_idx}"
                gene_gff_dict['gene'].append(gene_entry)

                starts.append(start)
                ends.append(end)
                chromosomes.append(chromosome)
                strands.append(strand)

        if not starts:  # 如果没有有效基因，跳过
            continue

        # 计算operon的范围
        chromosome = set(chromosomes).pop()
        operon_start = min(starts)
        operon_end = max(ends)
        strand = set(strands).pop()  # 默认正链，可根据需要修改逻辑
        genes_str = ",".join(genes)
        gene_entry = f"ID={formatted_idx};genes={genes_str}"

        operon_gff_dict['chromosome'].append(chromosome)
        operon_gff_dict['strand'].append(strand)
        operon_gff_dict['start'].append(operon_start)
        operon_gff_dict['end'].append(operon_end)
        operon_gff_dict['gene'].append(gene_entry)

    operon_gff_dict['source'] = ['SLRanger'] * len(operon_gff_dict['gene'])
    operon_gff_dict['type'] = ['operon'] * len(operon_gff_dict['gene'])
    operon_gff_dict['score'] = ['.'] * len(operon_gff_dict['gene'])
    operon_gff_dict['phase'] = ['.'] * len(operon_gff_dict['gene'])

    gene_gff_dict['source'] = ['SLRanger'] * len(gene_gff_dict['gene'])
    gene_gff_dict['type'] = ['gene'] * len(gene_gff_dict['gene'])
    gene_gff_dict['score'] = ['.'] * len(gene_gff_dict['gene'])
    gene_gff_dict['phase'] = ['.'] * len(gene_gff_dict['gene'])

    operon_gff_df = pd.DataFrame(operon_gff_dict)
    gene_gff_df = pd.DataFrame(gene_gff_dict)
    df = pd.concat([operon_gff_df, gene_gff_df])
    df_sorted = df.sort_values(by=['chromosome', 'start'])

    return df_sorted

def count_process(df, genes_dict):
    counts = df.value_counts().reset_index()
    counts_expand = fusion_expand(counts, genes_dict)
    counts_s = counts_expand.groupby(['gene', 'SL'])['count'].sum().reset_index()
    # has_semicolon = counts_filtered[counts_filtered['sort_fusion'].str.contains(';', na=False)]
    counts_re = reshape(counts_s)
    counts_re['sum_count'] = counts_re['SL2'] + counts_re['SL1']
    counts_re['sl2_ratio'] = counts_re['SL2'] / counts_re['sum_count']

    counts_re['type'] = counts_re.apply(lambda row: 'SL2' if row['sl2_ratio'] > 0.5 else 'SL1', axis=1)
    counts_re['type2'] = counts_re['type']
    # 满足条件的行赋值为 'SL2'
    counts_re.loc[(counts_re['SL2'] > 5) & (counts_re['sl2_ratio'] >= 0.25), 'type2'] = 'SL2'

    return counts_re

def main(args):
    print("Detection start ...")
    gff_file = args.gff
    df_genes = parse_gff(gff_file)
    df_genes_filter = parse_cds_gene(gff_file)
    df_genes_with_cds = pd.merge(df_genes, df_genes_filter, on='gene', how='right')
    df_pos_dict = df_genes_with_cds.set_index('gene').to_dict('index')
    df_pos = sort_and_calc_distance(df_genes_with_cds)
    args.mapping=run_track_cluster(args.gff,args.bam)
    map_gene = pd.read_csv(args.mapping, sep='\t', names=['query_name', 'gene'])
    sl_ss = sl_process(args.input, args.cutoff)
    sl_ss_gene = pd.merge(sl_ss, map_gene, how='left', on='query_name')
    counts_re = count_process(sl_ss_gene[['gene', 'SL']], df_pos_dict)
    # median_value_all = counts_re['sum_count'].median()
    median_value_sl2 = counts_re[counts_re['type2'] == 'SL2']['SL2'].median()
    counts_re = pd.merge(counts_re, df_pos, how='right', on='gene')
    counts_re['sum_count'] = counts_re['sum_count'].fillna(0)

    operon_result = group_genes_into_operons(counts_re, args.distance, median_value_sl2)
    updated_gene_list = merge_single_gene_sublists(operon_result, df_pos)

    # operon_combination = pd.DataFrame([','.join(sublist) for sublist in updated_gene_list])
    operon_combination_gff = generate_operon_gff(updated_gene_list, df_pos_dict)
    operon_combination_gff.to_csv(args.output, sep='\t', index=False, header=False)

    print('Operon detected to ' + args.output)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="help to know spliced leader and distinguish SL1 and SL2")
    parser.add_argument("-g", "--gff", type=str, required=True, help="GFF annotation file")
    parser.add_argument("-b", "--bam", type=str, required=True,help="bam file")
    parser.add_argument("-i", "--input", type=str, required=True, help="input the SL detection file")
    parser.add_argument("-o", "--output", type=str,  default="SLRanger.gff",help="output operon detection file")
    parser.add_argument("-d", "--distance", type=int, default=5000, help="promoter scope")
    parser.add_argument("-c", "--cutoff", type=float, default=4, help="cutoff of high confident SL sequence")
    args = parser.parse_args()
    main(args)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from plotnine import *
from plotnine.ggplot import PlotnineWarning
import re
import markdown
from pathlib import Path
from datetime import datetime
import seaborn as sns
import matplotlib.colors as mcolors
import warnings
# 隐藏特定的警告
warnings.filterwarnings('ignore', category=FutureWarning)
warnings.filterwarnings('ignore', category=DeprecationWarning)  # 如果使用plotnine可能需要
warnings.filterwarnings('ignore', category=PlotnineWarning)


# Function to calculate ratio
def sw_ratio(df, cols):
    df_long = pd.melt(df, value_vars=cols, var_name='group', value_name='score')
    df_counts = df_long.groupby(['score', 'group']).size().reset_index(name='count')
    # len_dict={
    #     'random':df_long[df_long['group']=='random'].shape[0],
    #     'SL_seq':df_long[df_long['group']=='SL_seq'].shape[0]
    # }
    # df_counts['count'] = df_counts.apply(lambda x: len_dict[x['group']]-x['count'], axis=1)
    df_wide = df_counts.pivot(index='score', columns='group', values='count').fillna(0)

    df_wide['ratio'] = df_wide[cols[1]] / df_wide[cols[0]]
    return df_wide

def plot_cumulative_line(data, output_name, cf):
    df_wide_sw = sw_ratio(data, ['random', 'SL_seq'])
    df_wide_sw.reset_index(inplace=True)
    # sw_sum = df_wide_sw['SL_seq'][df_wide_sw['ratio'] > 5].sum()
    sw_min = df_wide_sw['score'][df_wide_sw['ratio'] > cf].min()

    # Set ratios > 50 to infinity
    # df_wide_sw.loc[df_wide_sw['ratio'] > 50, 'ratio'] = np.inf
    # df_wide_sw['SL_seq'] = df_wide_sw['SL_seq'].sum() - df_wide_sw['SL_seq']
    # df_wide_sw['random'] = df_wide_sw['random'].sum() - df_wide_sw['random']
    # df_wide_sw['cumulative_sw'] = df_wide_sw['SL_seq'].cumsum()
    # df_wide_sw['cumulative_random'] = df_wide_sw['random'].cumsum()
    df_wide_sw['SL_seq'] = df_wide_sw['SL_seq'][::-1].cumsum()[::-1]
    df_wide_sw['random'] = df_wide_sw['random'][::-1].cumsum()[::-1]
    plot_df = pd.melt(df_wide_sw,id_vars=['score'],value_vars=['SL_seq', 'random'])

    # Create cumulative line plot
    tick_step = 5
    x_max = int(np.floor(plot_df['score'].max()))
    # 生成整数刻度（不超过 x_max）
    base_breaks = list(range(0, x_max + 1, tick_step))
    x_breaks = sorted(set(base_breaks + [sw_min]))
    p_cumulative = (
            ggplot(plot_df, aes(x='score', y='value', color='group'))
            + geom_line()
            + geom_point(size=1)
            + geom_vline(xintercept=sw_min, linetype='dashed', color='black')
            + scale_x_continuous(breaks=x_breaks)
            + labs(title='Cumulative Counts', x='Score', y='Cumulative Count')
            + theme_bw()
            + theme(
        plot_title=element_text(ha='center'),
        legend_position='right',
        axis_text_y=element_text(margin={'r': 10}),
    )
    )
    # Save the plot
    p_cumulative.save(output_name, width=5, height=3.5,format='png',dpi=300)

    return  sw_min

def plot_aligned_length(df,folder_name):

    def extract_number(s,unique_number):
        match = re.search(r'SL(\d+)', s)  # 使用正则表达式提取数字
        if s=='SL2_unknown':
            return unique_number - 1
        elif s =='SL1_unknown':
            return unique_number - 2
        if match:
            return int(match.group(1))  # 返回数字部分并转换为整数
        return 0  # 如果没有匹配到，返回 0（可以根据需要调整）
    # unknown will be merged together and SL1 separated
    # df['SL_type'].apply(lambda x: if 'unkown'(x))

    result_list=['SL1','SL2','SL3','SL4','SL5','SL6','SL7','SL8','SL9','SL10','SL11','SL12','SL13','SL1_unknown','SL2_unknown']
    colors_13 = [
        "#1F77B4",  # SL1 - 蓝色
        "#FF7F0E",  # SL2 - 橙色
        "#2CA02C",  # SL3 - 绿色
        "#D62728",  # SL4 - 红色
        "#9467BD",  # SL5 - 紫色
        "#8C564B",  # SL6 - 棕色
        "#E377C2",  # SL7 - 粉红
        "#00008B",  # SL8 - 深紫
        "#BCBD22",  # SL9 - 橄榄黄
        "#17BECF",  # SL10 - 青色
        "#AEC7E8",  # SL11 - 浅蓝
        "#FFBB78",  # SL12 - 浅橙
        "#98DF8A"  # SL13 - 浅绿
    ]

    # 创建完整的颜色映射字典
    color_mapping = {
        f"SL{i + 1}": colors_13[i] for i in range(13)
    }
    color_mapping['SL1_unknown']='#231815' # black
    color_mapping['SL2_unknown'] = 'darkgrey'
    df['SL_type'] = df['SL_type'].apply(lambda x: 'SL2_unknown' if 'unknown' in x and x != 'SL1_unknown' else x)
    data = df['SL_type'].value_counts().reset_index()
    category = pd.api.types.CategoricalDtype(categories=result_list, ordered=True)
    data['SL_type'] = data['SL_type'].astype(category)
    data = data.sort_values('SL_type')
    def autopct_func(pct, allvals):
        absolute = round(pct / 100. * sum(allvals))
        return f"{pct:.1f}%\n({absolute:d})"

    labels = data['SL_type'].values
    sizes = data['count'].values
    # 生成15个清新风格的颜色
    # 使用Seaborn的"husl"色系生成14个颜色，然后添加指定的灰色
    base_colors = sns.color_palette("pastel", len(labels)-1)  # 生成14个清新颜色
    # 将RGB颜色转换为HEX
    base_hex_colors = [mcolors.to_hex(color) for color in base_colors]
    # 添加最后一个指定的灰色

    plt.figure(figsize=(5, 5))
    plt.pie(sizes, labels=labels,
            colors=[color_mapping[label] for label in labels],
            pctdistance=0.65,  # 百分比标签距离中心的距离
            autopct=lambda pct: autopct_func(pct, sizes),
            startangle=90,
            textprops={'fontsize': 8},  # 调整字体大小
            labeldistance=1.1,  # 标签距离中心的距离
            rotatelabels=True)  # 旋转标签以避免重叠
    # 确保饼图为圆形
    plt.axis('equal')
    # 保存图形
    plt.savefig(folder_name+'type_pie.png', dpi=300, bbox_inches='tight')

    plot = ggplot(df, aes(x='query_length')) \
           + theme_bw()\
           + geom_histogram(binwidth=0.5) \
           + theme(
        figure_size=(8, 4),
        axis_text=element_text(size=12,),
        axis_title=element_text(size=12,),
        panel_grid_minor=element_blank(),
        title=element_text(size=12,),
        strip_background=element_rect(alpha=0),
        strip_text=element_text(size=12,),
        legend_position='bottom',
        legend_text=element_text(size=8,),
    )
    plot.save(folder_name+'query_length.png',dpi=300,format='png')

    Q1 = np.percentile(df['aligned_length'], 25)
    Q2 = np.percentile(df['aligned_length'], 50)
    Q3 = np.percentile(df['aligned_length'], 75)
    Q_list = [Q1, Q2, Q3]
    Q_axis_list = [[0, 0],
                   [1000, 1000],
                   [2000, 2000],
                   [3000, 3000],
                   ['Q1', Q_list[0]],
                   ['Q2', Q_list[1]],
                   ['Q3', Q_list[2]]]
    Q_axis_list = pd.DataFrame(Q_axis_list)
    df = df[df['aligned_length'] < 3500]
    Q_axis_list.sort_values(by=[1], inplace=True)
    plot = ggplot(df, aes(x='aligned_length')) \
           + theme_bw() \
           + geom_density(fill='#D7D7D7') \
           + scale_x_continuous(breaks=Q_axis_list[1].values.tolist(), labels=Q_axis_list[0].values.tolist()) \
           + theme(
        figure_size=(8, 4),
        axis_text=element_text(size=12, ),
        axis_title=element_text(size=12, ),
        panel_grid_minor=element_blank(),
        title=element_text(size=12, ),
        strip_background=element_rect(alpha=0),
        strip_text=element_text(size=12, ),
        legend_position='bottom',
        legend_text=element_text(size=8, ),
    )
    for item in Q_list:
        plot = plot + geom_vline(xintercept=item, color='black', alpha=0.5, linetype='dashdot')
    plot.save(folder_name + 'aligned_length.png', dpi=300, format='png')

    return data

def create_image_gallery_md_html(df, image_paths, output_md_path, output_html_path):
    # Markdown模板 - 添加表格部分
    md_template = """
# Data Visualization Results

This document contains the visualization results from the SLRanger.

## Data Summary Table

{data_table}

## Visualization Gallery

{image_sections}

---

*Generated on {current_date}. If you think this tool is pretty GOOD, don’t forget to give our [Git](https://github.com/lrslab/SLRanger) a star! *
    """

    # HTML样式模板 - 添加表格样式
    html_template = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Data Visualization Results</title>
    <style>
        body {{
            max-width: 1200px;
            margin: 0 auto;
            padding: 20px;
            line-height: 1.6;
        }}
        h1 {{
            color: #333;
            text-align: center;
            border-bottom: 2px solid #eee;
            padding-bottom: 10px;
        }}
        h2 {{
            color: #444;
            margin-top: 30px;
        }}
        img {{
            max-width: 100%;
            height: auto;
            display: block;
            margin: 20px auto;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
            border-radius: 4px;
        }}
        table {{
            width: 100%;
            border-collapse: collapse;
            margin: 20px 0;
        }}
        th, td {{
            padding: 8px;
            text-align: left;
            border: 1px solid #ddd;
        }}
        th {{
            background-color: #f5f5f5;
            color: #333;
        }}
        tr:nth-child(even) {{
            background-color: #fafafa;
        }}
        .footer {{
            text-align: center;
            color: #666;
            font-size: 0.9em;
            margin-top: 40px;
            padding-top: 20px;
            border-top: 1px solid #eee;
        }}
    </style>
</head>
<body>
    {html_content}
</body>
</html>
"""

    # 将DataFrame转换为Markdown表格
    data_table = df.to_markdown(index=False)

    # 创建每个图片的Markdown片段
    image_sections = ""
    image_captions = [
        "Cumulative Counts (SW). The **dashed black line** represents the recommended **cutoff** value.",
        "Cumulative Counts (SL). The **dashed black line** represents the recommended **cutoff** value.",
        "Query Length Distribution",
        "Aligned Length Distribution",
        "SL Type Distribution"
    ]

    for img_path, caption in zip(image_paths, image_captions):
        image_sections += f"""
### {caption}

![{caption}]({img_path})

---
        """

    # 获取当前日期
    current_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    # 生成完整的Markdown内容
    md_content = md_template.format(
        data_table=data_table,
        image_sections=image_sections,
        current_date=current_date
    )

    # 写入Markdown文件
    with open(output_md_path, 'w', encoding='utf-8') as f:
        f.write(md_content)

    # 将Markdown转换为HTML
    html_content = markdown.markdown(
        md_content,
        extensions=['markdown.extensions.extra', 'markdown.extensions.toc', 'markdown.extensions.tables']
    )

    # 将HTML内容嵌入到模板中
    final_html = html_template.format(html_content=html_content)

    # 写入HTML文件
    with open(output_html_path, 'w', encoding='utf-8') as f:
        f.write(final_html)

# 主程序

def visualize_html(output_file, cf):
    import os
    from datetime import datetime

    # 获取当前时间戳（格式：YYYYMMDD_HHMMSS）
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")


    path = output_file
    out_put_name = path.split('.')[0]

    # 创建文件夹名称（SLRange_view_中间加时间戳）
    folder_name = f"{out_put_name}_{timestamp}/"
    # 创建文件夹
    os.makedirs(folder_name, exist_ok=True)
    # Read data
    df = pd.read_csv(path, sep='\t')
    reads_all = len(df)
    df = df.dropna()
    reads_na = len(df)
    # print(reads_na / reads_all)

    # Data processing
    df['SL_score'] = df['SL_score'].astype(float)

    # Calculate AUC
    # SW processing
    data = df.copy()
    data['random'] = data['random_sw_score'].round(0)
    data['SL_seq'] = data['sw_score'].round(0)
    sw_min = plot_cumulative_line(data, folder_name+'cumulative_int_sw.png', cf)

    # SL processing
    data = df.copy()
    data['random'] = (data['random_SL_score'] * 2).round() / 2
    data['SL_seq'] = (data['SL_score'] * 2).round() / 2
    sl_min = plot_cumulative_line(data, folder_name+'cumulative_int_sl.png', cf)

    df = df[df['SL_type'] != 'random']
    reads_sw_solid = len(df[df['sw_score'] >= sw_min])
    potential_read = len(df)
    df = df[df['SL_score'] > sl_min]
    reads_sl_solid = len(df)
    # 图片文件路径（根据你的代码生成的图片名称）
    df['query_length'] = df['query_length'].astype(int).astype(str).astype(int)
    type_table = plot_aligned_length(df,folder_name)

    output_table = pd.DataFrame({
        'SL_type': ['Total', 'Candidate', 'Potential SL', 'SW Solid SL', 'SLRanger Solid SL'],
        'count': [reads_all, reads_na, potential_read, reads_sw_solid, reads_sl_solid]
    })

    output_table = pd.concat([output_table, type_table])
    output_table.columns = ['Variable', 'Read Count']
    output_table['Proportion to total reads (%)'] = output_table['Read Count'] / reads_all
    output_table['Proportion to Candidate reads (%)'] = output_table['Read Count'] / reads_na
    output_table['Proportion to Potential reads (%)'] = output_table['Read Count'] / potential_read
    output_table['Proportion to SLRanger Solid SL reads (%)'] = output_table['Read Count'] / reads_sl_solid

    proportion_cols = [col for col in output_table.columns if 'Proportion' in col]
    for col in proportion_cols:
        output_table[col] = (output_table[col] * 100).round(2)
        output_table.loc[output_table[col] >= 100, col] = '/'

    image_paths = [
        "cumulative_int.png",
        "cumulative_int_sl.png",
        "query_length.png",
        "aligned_length.png",
        "type_pie.png"
    ]

    # 输出文件路径
    output_md = folder_name+"visualization_results.md"
    output_html = folder_name+"visualization_results.html"
    output_table_path = folder_name+"summary_table.csv"
    # 创建Markdown和HTML文件

    output_table.to_csv(output_table_path, index=False)
    create_image_gallery_md_html(output_table,image_paths, output_md, output_html)
    print(f"Markdown file has been created at: {output_md}")
    print(f"HTML file has been created at: {output_html}")


import os
import time
from collections import OrderedDict
import subprocess
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
import warnings
from datetime import datetime
import shutil


def identify_file_path(file_path):
    if not os.path.exists(file_path):
        raise FileNotFoundError("File do not exist! Please check your path : " + file_path)

def run_cmd(cmd):
    try:
        # 执行命令并捕获输出
        print(cmd)
        subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT, universal_newlines=True)
        # 处理命令输出
        # print(output)
    except subprocess.CalledProcessError as e:
        # 处理命令执行错误
        print("Command execution failed:", cmd)
        print("Reason:", e.output)
        raise RuntimeError('There are some errors in the cmd as below, please check your env ')

def run_track_cluster(gff_file,bam_file):
    identify_file_path(gff_file)
    identify_file_path(bam_file)

    # 获取当前时间戳（格式：YYYYMMDD_HHMMSS）
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    # 创建文件夹名称（SLRange_view_中间加时间戳）
    folder_name = f"SLRange_temp_{timestamp}/"
    # 创建文件夹
    os.makedirs(folder_name, exist_ok=True)
    try:
        gff2bigg_cmd = 'gff2bigg.py -i '+gff_file+' -o '+folder_name+'ref.bed'
        run_cmd(gff2bigg_cmd)
        bam2bigg_cmd = 'bam2bigg.py -b '+bam_file+' -o '+folder_name+'read.bed'
        run_cmd(bam2bigg_cmd)
        bedtools_cmd = 'bedtools sort -i '+folder_name+'read.bed'+' >'+folder_name+'read_sort.bed'
        run_cmd(bedtools_cmd)
        track_cmd = "add_gene.py -r "+folder_name+'ref.bed'+" -s "+folder_name+"read_sort.bed"
        run_cmd(track_cmd)
    except Exception as e:
        print(e)
        # shutil.rmtree(folder_name)
        raise RuntimeError('There are some errors in the cmd as below, please check your env ')
    # shutil.rmtree(folder_name)
    return "read_sort_gene.bed"



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="SLRanger",
    version="0.0.0.8",
    author="GUO Zhihao",
    author_email="qhuozhihao@icloud.com",
    description='An integrated approach for spliced leader detection and operon prediction in eukaryotes using long RNA reads',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lrslab/SLRanger",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7.0,<=3.11.7',
    install_requires=[
        'bio>=1.5.0',
        'numpy>=1.23.0,<2.0.0',
        'pandas>=2.1.0',
        'plotnine>=0.12.4',
        'tqdm>=4.0.0',
        "pysam>=0.21.0",
        "biopython>=1.80",
        'pyssw==0.1.7',
        'trackcluster==0.1.7',
        "scikit-learn>=1.0.2",
        'tabulate>=0.8.0',
        'Markdown>=3.5',
        'seaborn>0.12.0'
    ],
    scripts=['SLRanger/SL_detect.py','SLRanger/operon_predict.py','SLRanger/add_gene.py']
)